(asdf:load-system 'hunchentoot)
(asdf:load-system 'dexador)
(asdf:load-system 'com.inuoe.jzon)
(asdf:load-system 'serapeum)
(asdf:load-system 'bordeaux-threads)
(asdf:load-system 'spinneret)
(asdf:load-system 'lass)
(asdf:load-system 'cl-smtp)
(asdf:load-system 'log4cl)
(asdf:load-system 'fiveam)
(asdf:load-system 'clack)
(asdf:load-system 'ningle)

(defpackage tfcserver
  (:use :cl :com.inuoe.jzon :spinneret :serapeum))

(in-package :tfcserver)

(log:config :daily "/tmp/tfc-%Y%m%d.log")

(defvar *app* (make-instance 'ningle:app))
(defconstant *nc-password* (uiop:getenv "NC-PASSWORD"))
(defconstant *email-password* (uiop:getenv "EMAIL-PASSWORD"))

(defparameter *mail-css*
  '((table :border-collapse "collapse" :width "100%")
    (("th," td) :padding "8px")
    (td :text-align "left" :width "70%")
    (th :text-align "right" :border-right "1px solid #ddd")
    (tr :border-bottom "1px solid #ddd")
    (h1 :text-align "center")))

(defun post-health-form (data-list)
  "Takes the health form as an alist and posts it to nextcloud"
  (let* ((student-name
           (concatenate 'string
                        (serapeum:trim-whitespace
                         (cdr (assoc "first-name" data-list
                                     :test 'string=)))
                        " "
                        (serapeum:trim-whitespace
                         (cdr (assoc "last-name" data-list
                                     :test 'string=)))))
         (parent-name
           (concatenate 'string
                        (serapeum:trim-whitespace
                         (cdr (assoc "parent-first-name" data-list
                                     :test 'string=)))
                        " "
                        (serapeum:trim-whitespace
                         (cdr (assoc "parent-last-name" data-list
                                     :test 'string=)))))
         (image (assoc "image" data-list))
         (hash-data (make-hash-table))
         (data `((37 . ,student-name)
                 (38 . ,parent-name)
                 (39 . ,(get-form-value "birth-date" data-list))
                 (40 . ,(get-form-value "street" data-list))
                 (41 . ,(get-form-value "city" data-list))
                 (42 . ,(get-form-value "state" data-list))
                 (43 . ,(get-form-value "zip" data-list))
                 (44 . ,(get-form-value "cell-phone" data-list))
                 (45 . ,(get-form-value "home-phone" data-list))
                 (46 . ,(get-form-value "additional-emergency-contact" data-list))
                 (47 . ,(get-form-value "doctor-name" data-list))
                 (48 . ,(get-form-value "doctor-city" data-list))
                 (49 . ,(get-form-value "doctor-phone" data-list))
                 (50 . ,(get-form-value "medical-coverage" data-list))
                 (51 . ,(get-form-value "insurance-name" data-list))
                 (52 . ,(get-form-value "policy-number" data-list))
                 (54 . ,(get-form-value "agreement" data-list))
                 (55 . ,(get-form-value "allergies" data-list))
                 (56 . ,(get-form-value "specific-allergies" data-list))
                 (57 . ,(get-form-value "allergic-treatment" data-list))
                 (58 . ,(get-form-value "conditions" data-list))
                 (59 . ,(get-form-value "tetanus-shot" data-list))
                 (60 . ,(get-form-value "medication-schedule" data-list))
                 (61 . ,(get-form-value "other-notes" data-list))
                 (62 . ,(get-form-value "swimming-ability" data-list)))))
    (loop for i in data
          do (setf (gethash (car i) hash-data) (cdr i)))
    (log:info data)
    (dex:post "https://staff.tfcconnection.org/ocs/v2.php/apps/tables/api/2/tables/4/rows"
              :basic-auth `("chris" . ,*nc-password*)
              :headers '(("OCS-APIRequest" . "true"))
              :content `(("data" . ,(stringify hash-data)))
              :verbose t)
    t))

(defun get-form-value (key list)
  "Takes the key and alist of an online form and returns the value from it"
  (serapeum:trim-whitespace (cdr (assoc key list :test 'string=))))


(defun post-camp-data (data)
  "Takes the camp data as an alist and sends it to nextcloud tables to be input"
  (let ((new-data `((63 . ,(concat
                            (get-form-value "first-name" data)
                            " "
                            (get-form-value "last-name" data)))
                    (64 . ,(concat
                            (get-form-value "parent-first-name" data)
                            " "
                            (get-form-value "parent-last-name" data)))
                    (65 . ,(get-form-value "parent-phone" data))
                    (66 . ,(get-form-value "parent-email" data))
                    (67 . ,(get-form-value "birth-date" data))
                    (69 . ,(get-form-value "gender" data))
                    (70 . ,(get-form-value "street" data))
                    (71 . ,(get-form-value "city" data))
                    (72 . ,(get-form-value "state" data))
                    (73 . ,(get-form-value "zip" data))
                    (74 . ,(get-form-value "grade" data))
                    (75 . ,(get-form-value "week" data))
                    (76 . ,(get-form-value "shirt" data))
                    (77 . ,(get-form-value "registration" data))))
        (hash-data (make-hash-table)))
    (loop for i in new-data
          do (setf (gethash (car i) hash-data) (cdr i)))
    (log:info hash-data)
    (bt:make-thread
     (lambda ()
       (dex:post "https://staff.tfcconnection.org/ocs/v2.php/apps/tables/api/2/tables/5/rows"
                 :basic-auth `("chris" . ,*nc-password*)
                 :headers '(("OCS-APIRequest" . "true"))
                 :content `(("data" . ,(stringify hash-data)))
                 :verbose t)))))

(defun mail-mt-form (form attachment)
  "Takes the form as an alist and sends a table formatted email
with the image attached to us"
  (log:info form)
  (let ((first-name (cdr (assoc "firstname" form :test 'string=)))
        (last-name (cdr (assoc "lastname" form :test 'string=)))
        (form (reverse form)))
    (not (cl-smtp:send-email
          "mail.tfcconnection.org"
          "no-reply@mail.tfcconnection.org"
          (if *devmode* '("chris@tfcconnection.org")
              '("chris@tfcconnection.org" "ethan@tfcconnection.org"))
          (format nil "~a ~a filled out a Mission Trip Form!" first-name last-name)
          (format nil "Mission Trip Form for ~a ~a" first-name last-name)
          :display-name "TFC ADMIN"
          :ssl :tls
          :authentication `(:login "no-reply@mail.tfcconnection.org" ,*email-password*)
          :attachments attachment
          :html-message 
          (with-html-string
            (:doctype)
            (:html
             (:head (:title "TFC Mission Trip Form")
                    (:style (apply #'lass:compile-and-write *mail-css*)))
             (:body
              (:h1 (format nil "Mission Trip Form for ~a ~a" first-name last-name))
              (:hr)
              (:table
               (loop for row in form
                     do (:tr
                         (:th (trim-whitespace (car row)))
                         (:td (trim-whitespace
                               (cdr row)))))))))))))

(defun mail-camp-form (form attachment)
  "Takes the form as an alist and sends a table formatted email
with the image attached"
  (let* ((first-name (trim-whitespace
                      (cdr (assoc "first-name" form
                                  :test 'string=))))
         (last-name (trim-whitespace
                     (cdr (assoc "last-name" form
                                 :test 'string=))))
         (parent-name (concatenate 'string
                                   (trim-whitespace
                                    (cdr (assoc "parent-first-name" form
                                                :test 'string=)))
                                   " "
                                   (trim-whitespace
                                    (cdr (assoc "parent-last-name" form
                                                :test 'string=))))))
    (log:info (format nil "Mailing out the camp form for ~a"
                      (concat first-name " " last-name)))
    (not (cl-smtp:send-email
          "mail.tfcconnection.org"
          "no-reply@mail.tfcconnection.org"
          (if *devmode* '("chris@tfcconnection.org")
              '("chris@tfcconnection.org" "ethan@tfcconnection.org"))
          (format nil "~a ~a filled out a Camp Form!" first-name last-name)
          (format nil "Camp Form for ~a ~a" first-name last-name)
          :display-name "TFC ADMIN"
          :ssl :tls
          :authentication `(:login "no-reply@mail.tfcconnection.org" ,*email-password*)
          :attachments attachment
          :html-message 
          (with-html-string
            (:doctype)
            (:html
             (:head (:title "TFC Health Form")
                    (:style (apply #'lass:compile-and-write *mail-css*)))
             (:body
              (:h1 (format nil "Camp Form for ~a ~a" first-name last-name))
              (:hr)
              (:table
               (loop for row in form
                     do (:tr
                         (:th (string-capitalize
                               (string-replace-all
                                "-" (car row) " ")))
                         (:td (cdr row))))))))))))

(defun mail-health-form (form attachment)
  "Takes the form as an alist and sends a table formatted email
with the image attached"
  (let* ((first-name (trim-whitespace
                      (cdr (assoc "first-name" form
                                  :test 'string=))))
         (last-name (trim-whitespace
                     (cdr (assoc "last-name" form
                                 :test 'string=))))
         (parent-name (concatenate 'string
                                   (trim-whitespace
                                    (cdr (assoc "parent-first-name" form
                                                :test 'string=)))
                                   " "
                                   (trim-whitespace
                                    (cdr (assoc "parent-last-name" form
                                                :test 'string=))))))
    (log:info (format nil "Mailing out the health form for ~a"
                      (concat first-name " " last-name)))
    (not (cl-smtp:send-email
          "mail.tfcconnection.org"
          "no-reply@mail.tfcconnection.org"
          (if *devmode* '("chris@tfcconnection.org")
              '("chris@tfcconnection.org" "ethan@tfcconnection.org"))
          (format nil "~a ~a filled out a Health Form!" first-name last-name)
          (format nil "Health Form for ~a ~a" first-name last-name)
          :display-name "TFC ADMIN"
          :ssl :tls
          :authentication `(:login "no-reply@mail.tfcconnection.org" ,*email-password*)
          :attachments attachment
          :html-message 
          (with-html-string
            (:doctype)
            (:html
             (:head (:title "TFC Health Form")
                    (:style (apply #'lass:compile-and-write *mail-css*)))
             (:body
              (:h1 (format nil "Health Form for ~a ~a" first-name last-name))
              (:hr)
              (:table
               (loop for row in form
                     do (:tr
                         (:th (string-capitalize
                               (string-replace-all
                                "-" (car row) " ")))
                         (:td (cdr row))))))))))))

(setf (ningle:route *app* "/mt-form" :method :post)
      #'(lambda (params)
          (log:info params)))

(setf (ningle:route *app* "/health-form" :method :post)
      #'(lambda (params)
          (log:info params)
          (let* ((image-element (assoc "image" params :test 'string=))
                 (registration (cdr (assoc "registration" params :test 'string=)))
                 (first-name (cdr (assoc "first-name" params :test 'string=)))
                 (last-name (cdr (assoc "last-name" params :test 'string=)))
                 (image (nth 1 image-element))
                 (filename (nth 2 image-element)))
            (serapeum:write-stream-into-file
             image (pathname "frodo.jpg")
             :if-exists :supersede
             :if-does-not-exist :create)
            (log:info filename)
            (bt:make-thread
             (lambda ()
               (post-health-form params)
               (log:info "Sent to Nextcloud")
               (mail-health-form params (pathname filename))
               (log:info "Sent email"))
             :name "health-form")
            (setf (lack.response:response-headers ningle:*response*)
                  (append (lack.response:response-headers ningle:*response*)
                          (list :access-control-expose-headers "*")))
            (log:info registration)
            (cond ((string= registration "now")
                   (log:info "Sending them to pay now")
                   (setf (lack.response:response-headers ningle:*response*)
                         (append (lack.response:response-headers ningle:*response*)
                                 (list :hx-redirect
                                       "https://secure.myvanco.com/L-Z772/campaign/C-13JPJ"))))
                  ((string= registration "full")
                   (log:info "Sending them to pay full amount")
                   (setf (lack.response:response-headers ningle:*response*)
                         (append (lack.response:response-headers ningle:*response*)
                                 (list :hx-redirect
                                       "https://secure.myvanco.com/L-Z772/campaign/C-13JQE"))))
                  (t
                   (log:info "Sending the health form thank you snippet")
                   (with-html-string
                     (:div
                      :class "mt-8"
                      (:h2 (format nil
                                   "Thank You ~A!"
                                   (concat
                                    first-name " " last-name)))
                      (:p :class "text-md"
                          "If you'd like to pay for your registration go to the donate tab in the top right when you are ready and find the camp registration option."))))))))

(hunchentoot:define-easy-handler (camp-form :uri "/camp-form") ()
  (let* ((request-type (hunchentoot:request-method hunchentoot:*request*))
         (data (hunchentoot:post-parameters* hunchentoot:*request*)))
    (camp-form data)))

(defun camp-form (data)
  "Process the camp form"
  (log:info data)
  ;; This is extremely necessary so that cors is right
  (let ((registration (cdr (assoc "registration" data :test 'string=)))
        (health (cdr (assoc "health-form" data :test 'string=))))
    (when (boundp 'hunchentoot:*reply*)
      (setf (tbnl:header-out :access-control-expose-headers) "*"))
    (when data 
      (post-camp-data data)
      (mail-camp-form data nil))
    (if (string= health "later")
        (progn (uiop:println "Selected health later")
               (cond ((string= registration "now")
                      (when (boundp 'tbnl:*reply*)
                        (setf (hunchentoot:header-out :HX-Redirect) "https://secure.myvanco.com/L-Z772/campaign/C-13JPJ")
                        (log:info (tbnl:headers-out*)))
                      (log:info "Sending them to pay now")
                      "Redirecting to paying now")
                     ((string= registration "full")
                      (when (boundp 'tbnl:*reply*)
                        (setf (tbnl:header-out :HX-Redirect) "https://secure.myvanco.com/L-Z772/campaign/C-13JQE")
                        (log:info (tbnl:headers-out*)))
                      (log:info "Sending them to pay full amount")
                      "Redirecting to paying full amount")
                     ((string= registration "later")
                      (let ((first-name (cdr (assoc "first-name" data :test 'string=)))
                            (last-name (cdr (assoc "last-name" data :test 'string=))))
                        (log:info "Sending the camp thank you snippet")
                        (with-html-string
                          (:div
                           :class "mt-8"
                           (:h2 (format nil
                                        "Thank You ~A!"
                                        (concat
                                         first-name " " last-name)))
                           (:p "Can't wait to see you at camp!!")
                           (:p :class "text-md"
                               "If you'd like to pay for your registration go to the donate tab in the top right when you are ready and find the camp registration option.")))))))
        (when (string= health "now")
          (when (boundp 'tbnl:*reply*)
            (setf (tbnl:header-out :HX-Redirect) (format nil "/camp-health-form/?registration=~A" registration))
            (log:info (tbnl:headers-out*)))
          (log:info "Sending them to the health form for camp")
          "Redirecting to the health form"))))

(defun main ()
  (clack:clackup *app* :port 4242)
  (log:info "Server has started on port 4242~&")
  (handler-case
      (bt:join-thread (find-if (lambda (th)
                                 (search "hunchentoot" (bt:thread-name th)))
                               (bt:all-threads)))
    (#+sbcl sb-sys:interactive-interrupt
     #+ccl  ccl:interrupt-signal-condition
     #+clisp system::simple-interrupt-condition
     #+ecl ext:interactive-interrupt
     #+allegro excl:interrupt-signal
     () (progn
          (format *error-output* "Aborting.~&")
          (hunchentoot:stop *server*)
          (uiop:quit)))
    (error (c) (format t "Woops, an unknown error occured:~&~a~&" c))))

(setf test-data '(("first-name" . "Frodo") ("last-name" . "Braggins")
                    ("parent-first-name" . "Bilbo")
                    ("parent-last-name" . "Braggins")
                    ("birth-date" . "1857-04-06") ("gender" . "Male")
                    ("street" . "341 West Hobbiton") ("city" . "The Shire")
                    ("state" . "Middle Earth") ("zip" . "88888") ("grade" . "freshman")
                    ("parent-phone" . "9998887777")
                    ("parent-email" . "bilbosmells@theshire.com")
                    ("allergies" . "No") ("week" . "week1")
                    ("shirt" . "medium") ("final-agreement" . "yes")
                    ("health-form" . "now") ("registration" . "now")))

(fiveam:test test-camp-form
  (fiveam:is (string= "Redirecting to the health form" (camp-form test-data))))

(fiveam:test test-post-health-form
  (fiveam:is (post-health-form test-data)))
